////////////////////////////////////////////////////////////////////////
//
//   Harvard Computer Science
//   CS 175: Computer Graphics
//   Professor Steven Gortler
//
////////////////////////////////////////////////////////////////////////

// I N C L U D E S /////////////////////////////////////////////////////

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <vector>

#include <GL/glew.h>
#ifdef __MAC__
#   include <GLUT/glut.h>
#else
#   include <GL/glut.h>
#endif
#include "matrix4.h"
#include "ppm.h"
#include "glsupport.h"

using namespace std;

// G L O B A L S ///////////////////////////////////////////////////////

// application globals
static int g_width             = 512;         // screen width
static int g_height            = 512;         // screen height


static const float g_frustMinFov = 60.0; 
static const float g_frustNear = -0.1;   
static const float g_frustFar = -50.0;   

static const double floor_size = 10.0;	
static const double floor_y = -2.0;	


static bool g_leftClicked       = false;    
static bool g_rightClicked      = false;    
static int g_leftClickX, g_leftClickY;

static int g_rightClickX, g_rightClickY;    


// OpenGL non-shader handles
static GLuint h_floorVertVbo,h_floorNormVbo,
  h_cubeVertVbo,h_cubeNormVbo;

// Shader globals
static GLuint h_program;     // handle to OpenGL program object

// handles to vertex attributes (note the `a' prefix)
static GLint h_aVertex;
static GLint h_aNormal;
static GLint h_aColor;
static GLint h_uLight;
static GLint h_uProjMatrix;
static GLint h_uModelViewMatrix;
static GLint h_uNormalMatrix;

static const Cvec4 g_light(2.0, 3.0, 14.0, 1.);

static Matrix4 g_eyeRbt = Matrix4::makeTranslation(Cvec3(0.0, 0., 4.0));
static Matrix4 g_objRbt = Matrix4::makeTranslation(Cvec3(-1,0,-1))
  *Matrix4::makeXRotation(22.0);; 



static void drawObj(GLuint vertbo, GLuint normbo,  int numverts) {
  safe_glEnableVertexAttribArray(h_aVertex);
  safe_glEnableVertexAttribArray(h_aNormal);


  glBindBuffer(GL_ARRAY_BUFFER,vertbo);
  safe_glVertexAttribPointer(h_aVertex, 3, GL_FLOAT, GL_FALSE, 0, 0);

  glBindBuffer(GL_ARRAY_BUFFER,normbo);
  safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);


  glDrawArrays(GL_TRIANGLES,0,numverts);

  safe_glDisableVertexAttribArray(h_aVertex);
  safe_glDisableVertexAttribArray(h_aNormal);
}


// C A L L B A C K S ///////////////////////////////////////////////////


static void sendProjectionMatrix(const Matrix4& projMatrix) {
  GLfloat glmatrix[16];
  projMatrix.writeToColumnMajorMatrix(glmatrix); // send projection matrix
  safe_glUniformMatrix4fv(h_uProjMatrix, glmatrix);
}

// takes MVM and its normal matrix to the shaders
static void sendModelViewNormalMatrix(const Matrix4& MVM, const Matrix4& NMVM) {
  GLfloat glmatrix[16];
  MVM.writeToColumnMajorMatrix(glmatrix); // send MVM
  safe_glUniformMatrix4fv(h_uModelViewMatrix, glmatrix);

  NMVM.writeToColumnMajorMatrix(glmatrix); // send NMVM
  safe_glUniformMatrix4fv(h_uNormalMatrix, glmatrix);
}


static void display(void) {
  glUseProgram(h_program);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


  const Matrix4 projmat = Matrix4::makeProjection(g_frustMinFov, 
						  g_width/ static_cast <double> (g_height), 
						  g_frustNear, g_frustFar);
  sendProjectionMatrix(projmat);
  
  Matrix4 MVM = inv(g_eyeRbt);
  Matrix4 NMVM = normalMatrix(MVM);
  sendModelViewNormalMatrix(MVM, NMVM);


  Cvec4 lightE = inv(g_eyeRbt) * g_light;
  safe_glUniform3f(h_uLight, lightE[0], lightE[1], lightE[2]);
  

  safe_glVertexAttrib3f(h_aColor, .6, .8, .6);
  drawObj(h_floorVertVbo,h_floorNormVbo,6);

  MVM = inv(g_eyeRbt) *g_objRbt;
  NMVM = normalMatrix(MVM);
  sendModelViewNormalMatrix(MVM, NMVM);

  safe_glVertexAttrib3f(h_aColor, 0, 0, 1);
  drawObj(h_cubeVertVbo,h_cubeNormVbo,36);
 
  glutSwapBuffers();
  checkGlErrors();
}


// _____________________________________________________
//|                                                     |
//|  reshape                                            |
//|_____________________________________________________|
///
///  Whenever a window is resized, a "resize" event is
///  generated and glut is told to call this reshape
///  callback function to handle it appropriately.

static void reshape(int w, int h) {
  g_width = w;
  g_height = h;
  glViewport(0, 0, w, h);
  glutPostRedisplay();
}


static void mouse(int button, int state, int x, int y) {
  if (button == GLUT_LEFT_BUTTON) {
    if (state == GLUT_DOWN) {
      // right mouse button has been clicked
      g_leftClicked = true;
      g_leftClickX = x;
      g_leftClickY = g_height - y - 1;
    }
    else {
      // right mouse button has been released
      g_leftClicked = false;
    }
  }
}


static void motion(int x, int y) {
  const int newx = x;
  const int newy = g_height - y - 1;
  if (g_leftClicked) {
    float deltax = (newx - g_leftClickX) * 0.02;

    
    g_objRbt = g_objRbt * Matrix4::makeYRotation(10*deltax);
    // this just shows that you can move the cube.
    // see the book for a description of the preferred ways to move the eye
    // and objects.


    g_leftClickX = newx;
    g_leftClickY = newy;
  }
  glutPostRedisplay();
}



// H E L P E R    F U N C T I O N S ////////////////////////////////////


static void initGlutState(int argc, char **argv) {
  glutInit(&argc,argv); 
  glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH);
  glutInitWindowSize(g_width, g_height);   
  glutCreateWindow("Hello World"); 
  glutMotionFunc(motion);                  
  glutMouseFunc(mouse);                    
  glutDisplayFunc(display);                
  glutReshapeFunc(reshape);                
}

static void initGLState() {
  glClearColor(128./255,200./255,1,0);
  glEnable(GL_FRAMEBUFFER_SRGB);

  glClearDepth(0.);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_GREATER);
  glCullFace(GL_BACK);
  glReadBuffer(GL_BACK);
}

static void initShaders() {
  h_program = glCreateProgram();
  readAndCompileShader(h_program, "./shaders/asst1-gl3.vshader", 
		       "./shaders/asst1-gl3.fshader");

  glBindFragDataLocation(h_program, 0, "fragColor");

  // grab handles to the shader variables by name

  h_uLight = safe_glGetUniformLocation(h_program, "uLight");
  h_uProjMatrix = safe_glGetUniformLocation(h_program, "uProjMatrix");
  h_uModelViewMatrix = safe_glGetUniformLocation(h_program, "uModelViewMatrix");
  h_uNormalMatrix = safe_glGetUniformLocation(h_program, "uNormalMatrix");
  
  // Retrieve handles to vertex attributes
  h_aVertex = safe_glGetAttribLocation(h_program, "aVertex");
  h_aNormal = safe_glGetAttribLocation(h_program, "aNormal");
  h_aColor = safe_glGetAttribLocation(h_program, "aColor");
  

}

static void initVBOs() {
GLfloat floorVerts[18] = 
  {  
    -floor_size, floor_y, -floor_size,
    floor_size, floor_y, floor_size,
    floor_size, floor_y, -floor_size,
    -floor_size, floor_y, -floor_size,
    -floor_size, floor_y, floor_size,
    floor_size, floor_y, floor_size
  };
GLfloat floorNorms[18] = 
  {  
    0,1,0,
    0,1,0,
    0,1,0,
    0,1,0,
    0,1,0,
    0,1,0
  };

GLfloat cubeVerts[36*3]=
  { 
    -0.5f, -0.5f, -0.5f,
    -0.5f, -0.5f, +0.5f,
    +0.5f, -0.5f, +0.5f,
    
    -0.5f, -0.5f, -0.5f,
    +0.5f, -0.5f, +0.5f,
    +0.5f, -0.5f, -0.5f,

    -0.5f, 0.5f, -0.5f, 
    -0.5f, 0.5f, +0.5f, 
    +0.5f, 0.5f, +0.5f, 
    
    -0.5f, 0.5f, -0.5f, 
    +0.5f, 0.5f, +0.5f, 
    +0.5f, 0.5f, -0.5f, 


     -0.5f, -0.5f, -0.5f,
     -0.5f, +0.5f, -0.5f,
     +0.5f, +0.5f, -0.5f,

     -0.5f, -0.5f, -0.5f,
     +0.5f, +0.5f, -0.5f,
     +0.5f, -0.5f, -0.5f,

     -0.5f, -0.5f, 0.5f, 
     -0.5f, +0.5f, 0.5f, 
     +0.5f, +0.5f, 0.5f, 

     -0.5f, -0.5f, 0.5f, 
     +0.5f, +0.5f, 0.5f, 
     +0.5f, -0.5f, 0.5f, 


     -0.5f, -0.5f, -0.5f,
     -0.5f, -0.5f, +0.5f,
     -0.5f, +0.5f, +0.5f,

     -0.5f, -0.5f, -0.5f,
     -0.5f, +0.5f, +0.5f,
     -0.5f, +0.5f, -0.5f,


     0.5f, -0.5f, -0.5f, 
     0.5f, -0.5f, +0.5f, 
     0.5f, +0.5f, +0.5f, 

     0.5f, -0.5f, -0.5f, 
     0.5f, +0.5f, +0.5f, 
     0.5f, +0.5f, -0.5f


  };

// Normals of a cube.
GLfloat cubeNorms[36*3] = 
  { +0.0f, -1.0f, +0.0f,
    +0.0f, -1.0f, +0.0f,
    +0.0f, -1.0f, +0.0f,
    +0.0f, -1.0f, +0.0f,
    +0.0f, -1.0f, +0.0f,
    +0.0f, -1.0f, +0.0f,

    +0.0f, +1.0f, +0.0f,
    +0.0f, +1.0f, +0.0f,
    +0.0f, +1.0f, +0.0f,
    +0.0f, +1.0f, +0.0f,
    +0.0f, +1.0f, +0.0f,
    +0.0f, +1.0f, +0.0f,

    +0.0f, +0.0f, -1.0f,
    +0.0f, +0.0f, -1.0f,
    +0.0f, +0.0f, -1.0f,
    +0.0f, +0.0f, -1.0f,
    +0.0f, +0.0f, -1.0f,
    +0.0f, +0.0f, -1.0f,

    +0.0f, +0.0f, +1.0f,
    +0.0f, +0.0f, +1.0f,
    +0.0f, +0.0f, +1.0f,
    +0.0f, +0.0f, +1.0f,
    +0.0f, +0.0f, +1.0f,
    +0.0f, +0.0f, +1.0f,

    -1.0f, +0.0f, +0.0f,
    -1.0f, +0.0f, +0.0f,
    -1.0f, +0.0f, +0.0f,
    -1.0f, +0.0f, +0.0f,
    -1.0f, +0.0f, +0.0f,
    -1.0f, +0.0f, +0.0f,

    +1.0f, +0.0f, +0.0f,
    +1.0f, +0.0f, +0.0f,
    +1.0f, +0.0f, +0.0f,
    +1.0f, +0.0f, +0.0f,
    +1.0f, +0.0f, +0.0f,
    +1.0f, +0.0f, +0.0f
  };



  glGenBuffers(1,&h_floorVertVbo);
  glBindBuffer(GL_ARRAY_BUFFER,h_floorVertVbo);
  glBufferData(
    GL_ARRAY_BUFFER,
    18*sizeof(GLfloat),
    floorVerts,
    GL_STATIC_DRAW);


  glGenBuffers(1,&h_floorNormVbo);
  glBindBuffer(GL_ARRAY_BUFFER,h_floorNormVbo);
  glBufferData(
    GL_ARRAY_BUFFER,
    18*sizeof(GLfloat),
    floorNorms,
    GL_STATIC_DRAW);



  glGenBuffers(1,&h_cubeVertVbo);
  glBindBuffer(GL_ARRAY_BUFFER,h_cubeVertVbo);
  glBufferData(
    GL_ARRAY_BUFFER,
    36*3*sizeof(GLfloat),
    cubeVerts,
    GL_STATIC_DRAW);


  glGenBuffers(1,&h_cubeNormVbo);
  glBindBuffer(GL_ARRAY_BUFFER,h_cubeNormVbo);
  glBufferData(
    GL_ARRAY_BUFFER,
    36*3*sizeof(GLfloat),
    cubeNorms,
    GL_STATIC_DRAW);
}



// M A I N /////////////////////////////////////////////////////////////

// _____________________________________________________
//|                                                     |
//|  main                                               |
//|_____________________________________________________|
///
///  The main entry-point for the HelloWorld example
///  application.

int main(int argc, char **argv) {

  initGlutState(argc,argv);
  glewInit();  
  
  if (!GLEW_VERSION_3_0) {
    cerr << "Error: card/driver does not support OpenGL Shading Language v1.3\n";
    return -1;
  }

  initGLState();
  initShaders();
  initVBOs();

  glutMainLoop();

  return 0;
}
